use std::fmt::{Display, Formatter, Result as FmtResult};
use std::io::{self, Write};
use std::thread;
use std::time::{Duration, Instant};

#[derive(Debug, Copy, Clone, Default)]
pub struct Pomodoro {
    state: State,
    count: u64,
    wait_start: Option<Instant>,
    paused: PausedState,
}

static POLLING_RATE: Duration = Duration::from_millis(300);
static WORK_TIME: u64 = 2;
static SBREAK_TIME: u64 = 2;
static LBREAK_TIME: u64 = 2;

impl Pomodoro {
    pub fn new() -> Self {
        Self::default()
    }

    pub fn start(&mut self) {
        loop {
            if let Status::Complete = self.poll() {
                break;
            }
            thread::sleep(POLLING_RATE);
        }
    }

    fn poll(&mut self) -> Status {
        let status = self.check();

        match status {
            Status::Paused => {
                assert!(self.paused.value);
                let now = Instant::now();

                self.paused.duration += match self.paused.previous {
                    Some(earlier) => now - earlier,
                    None => Default::default(),
                };

                self.paused.previous = Some(now);
            }
            Status::Active => self.display_time(),
            Status::NextState => {
                let (update_count, new_state) = self.next();

                if let Count::Increase = update_count {
                    self.count += 1;
                }

                self.state = new_state;
                self.wait_start = match new_state {
                    State::Inactive => None,
                    State::Work | State::ShortBreak | State::LongBreak => Some(Instant::now()),
                };

                self.display_time();
            }
            Status::Complete => {
                println!("\rPomodoro Cycle Complete!");
            }
            Status::Inactive => {}
        };

        status
    }

    fn display_time(&self) {
        if let Some(earlier) = self.wait_start {
            let wait = Instant::now() - earlier;
            let left: Clock = (Self::wait_times(self.state) - wait).into();
            let _ = Self::print(&format!("\r{} {}", self.state, left));
        }
    }

    fn print(text: &str) -> io::Result<()> {
        let out = io::stdout();
        let mut handle = out.lock();
        handle.write_all(text.as_bytes())?;
        handle.flush()?;
        Ok(())
    }

    fn next(&self) -> (Count, State) {
        match self.state {
            State::Work => {
                let state: State = if (self.count + 1) % 4 == 0 {
                    State::LongBreak
                } else {
                    State::ShortBreak
                };

                (Count::Increase, state)
            }
            State::ShortBreak | State::Inactive => (Count::Remain, State::Work),
            State::LongBreak => (Count::Remain, State::Inactive),
        }
    }

    fn check(&self) -> Status {
        if self.paused.value {
            return Status::Paused;
        }

        match self.wait_start {
            Some(earlier) => {
                let diff: Duration = (Instant::now() - self.paused.duration) - earlier;

                if diff > Self::wait_times(self.state) {
                    Status::NextState
                } else {
                    Status::Active
                }
            }
            None => {
                if self.count == 0 {
                    Status::NextState
                } else {
                    Status::Complete
                }
            }
        }
    }

    fn wait_times(state: State) -> Duration {
        match state {
            State::Work => Duration::from_secs(WORK_TIME),
            State::ShortBreak => Duration::from_secs(SBREAK_TIME),
            State::LongBreak => Duration::from_secs(LBREAK_TIME),
            _ => unreachable!("Can not have Pomodoro state = State::Inactive and wait_start = Some(...) at the same time.")
        }
    }
}

struct Clock {
    hours: u64,
    minutes: u64,
    seconds: u64,
}

impl From<Duration> for Clock {
    fn from(dur: Duration) -> Self {
        let dur = dur.as_secs();

        let hours = dur / 3600;
        let minutes = (dur - (hours * 3600)) / 60;
        let seconds = dur - (hours * 3600) - (minutes * 60);

        Self {
            hours,
            minutes,
            seconds,
        }
    }
}

impl Display for Clock {
    fn fmt(&self, f: &mut Formatter<'_>) -> FmtResult {
        if self.hours > 0 {
            // No need to print the hours if there are none.
            if self.hours < 10 {
                write!(f, "0{}:", self.hours)?;
            } else {
                write!(f, "{}:", self.hours)?;
            }
        }

        if self.minutes < 10 {
            write!(f, "0{}:", self.minutes)?;
        } else {
            write!(f, "{}:", self.minutes)?;
        }

        if self.seconds < 10 {
            write!(f, "0{}", self.seconds)
        } else {
            write!(f, "{}", self.seconds)
        }
    }
}

#[derive(Debug, Copy, Clone, Default)]
struct PausedState {
    value: bool,
    duration: Duration,
    previous: Option<Instant>,
}

#[derive(Debug, Copy, Clone)]
enum Count {
    Increase,
    Remain,
}

#[derive(Debug, Copy, Clone)]
enum Status {
    Paused,
    Active,
    NextState,
    Inactive,
    Complete,
}

#[derive(Debug, Copy, Clone)]
enum State {
    Inactive,
    Work,
    ShortBreak,
    LongBreak,
}

impl Default for State {
    fn default() -> Self {
        State::Inactive
    }
}

impl Display for State {
    fn fmt(&self, f: &mut Formatter<'_>) -> FmtResult {
        match *self {
            State::Work => f.write_str("Working: "),
            State::ShortBreak => f.write_str("Resting: "),
            State::LongBreak => f.write_str("REALLY Resting: "),
            State::Inactive => f.write_str("Inactive: "),
        }
    }
}
